@@ -6,7 +6,9 @@
 #include <span>
 
 #include "glaze/api/std/span.hpp"
+#include "glaze/core/buffer_traits.hpp"
 #include "glaze/core/common.hpp"
+#include "glaze/core/streaming_state.hpp"
 #include "glaze/util/parse.hpp"
 
 namespace glz
@@ -39,15 +41,17 @@ namespace glz
       if constexpr (Opts.format != NDJSON) {
          if (buffer.empty()) [[unlikely]] {
             ctx.error = error_code::no_read_input;
-            return {ctx.error, ctx.custom_error_message, 0, ctx.includer_error};
+            return {0, ctx.error, ctx.custom_error_message};
          }
       }
 
-      constexpr bool use_padded = resizable<Buffer> && non_const_buffer<Buffer> && !has_disable_padding(Opts);
+      constexpr bool use_padded = resizable<Buffer> && non_const_buffer<Buffer> && !check_disable_padding(Opts);
 
+      [[maybe_unused]] size_t original_size{};
       if constexpr (use_padded) {
          // Pad the buffer for SWAR
-         buffer.resize(buffer.size() + padding_bytes);
+         original_size = buffer.size();
+         buffer.resize(original_size + padding_bytes);
       }
 
       auto [it, end] = read_iterators<Opts, use_padded>(buffer);
@@ -100,10 +104,10 @@ namespace glz
 
       if constexpr (use_padded) {
          // Restore the original buffer state
-         buffer.resize(buffer.size() - padding_bytes);
+         buffer.resize(original_size);
       }
 
-      return {ctx.error, ctx.custom_error_message, size_t(it - start), ctx.includer_error};
+      return {size_t(it - start), ctx.error, ctx.custom_error_message};
    }
 
    template <auto Opts, class T>
@@ -127,7 +131,7 @@ namespace glz
    {
       const auto str = std::string_view{std::forward<Buffer>(buffer)};
       if (str.empty()) {
-         return {error_code::no_read_input};
+         return {0, error_code::no_read_input};
       }
       return read<Opts>(value, str, ctx);
    }
@@ -139,4 +143,64 @@ namespace glz
       context ctx{};
       return read<Opts>(value, std::forward<Buffer>(buffer), ctx);
    }
+
+   // Streaming read for input streaming buffers
+   // Uses incremental parsing with internal refill points for arrays/objects
+   // Returns error_ctx with total bytes consumed across all refills
+   template <auto Opts, class T, class Buffer, has_streaming_state Ctx>
+      requires read_supported<T, Opts.format> && is_input_streaming<std::remove_reference_t<Buffer>>
+   [[nodiscard]] error_ctx read_streaming(T& value, Buffer&& buffer, Ctx&& ctx)
+   {
+      // For streaming, we need null_terminated = false to track indentation_level
+      static constexpr auto StreamingOpts = [] {
+         auto o = is_padded_off<Opts>();
+         o.null_terminated = false;
+         return o;
+      }();
+
+      // Initial fill if buffer is empty
+      if (buffer.empty()) {
+         if (!refill_buffer(buffer)) {
+            if constexpr (Opts.format != NDJSON) {
+               ctx.error = error_code::no_read_input;
+               return {0, ctx.error, ctx.custom_error_message};
+            }
+         }
+      }
+
+      // Set up streaming state so parsers can trigger internal refills
+      ctx.stream = make_streaming_state(buffer);
+
+      auto [it, end] = read_iterators<Opts>(buffer);
+
+      // Parse with streaming-aware context
+      // The parser will internally refill as needed at safe points (between array elements, etc.)
+      parse<Opts.format>::template op<StreamingOpts>(value, ctx, it, end);
+
+      // Calculate final consumed bytes
+      // Note: During streaming, the buffer may have been refilled multiple times,
+      // so we use bytes_consumed() which tracks total consumption
+      const size_t final_consumed = static_cast<size_t>(it - ctx.stream.data());
+      consume_buffer(buffer, final_consumed);
+
+      // Handle end_reached as success when parsing completed at depth 0
+      // (same logic as non-streaming read)
+      if (ctx.error == error_code::end_reached && ctx.indentation_level == 0) {
+         ctx.error = error_code::none;
+      }
+
+      if (bool(ctx.error)) {
+         return {buffer.bytes_consumed(), ctx.error, ctx.custom_error_message};
+      }
+
+      return {buffer.bytes_consumed(), error_code::none, ctx.custom_error_message};
+   }
+
+   template <auto Opts, class T, class Buffer>
+      requires read_supported<T, Opts.format> && is_input_streaming<std::remove_reference_t<Buffer>>
+   [[nodiscard]] error_ctx read_streaming(T& value, Buffer&& buffer)
+   {
+      streaming_context ctx{};
+      return read_streaming<Opts>(value, std::forward<Buffer>(buffer), ctx);
+   }
 }
